from __future__ import annotations

from pathlib import Path
from unittest.mock import patch
from urllib.parse import urlparse

from murfey.client.contexts.tomo import TomographyContext
from murfey.client.instance_environment import MurfeyInstanceEnvironment


def test_tomography_context_initialisation_for_tomo(tmp_path):
    context = TomographyContext("tomo", tmp_path)
    assert not context._last_transferred_file
    assert context._acquisition_software == "tomo"


@patch("requests.get")
@patch("requests.post")
def test_tomography_context_add_tomo_tilt(mock_post, mock_get, tmp_path):
    env = MurfeyInstanceEnvironment(
        url=urlparse("http://localhost:8000"),
        client_id=0,
        sources=[tmp_path],
        default_destinations={tmp_path: str(tmp_path)},
    )
    context = TomographyContext("tomo", tmp_path)
    (tmp_path / "Position_1_[30.0]_fractions.tiff").touch()
    context.post_transfer(
        tmp_path / "Position_1_[30.0]_fractions.tiff",
        role="detector",
        required_position_files=[],
        required_strings=["fractions"],
        environment=env,
    )
    assert context._tilt_series == {
        "Position_1": [tmp_path / "Position_1_[30.0]_fractions.tiff"]
    }
    assert (
        context._last_transferred_file == tmp_path / "Position_1_[30.0]_fractions.tiff"
    )
    (tmp_path / "Position_1_[-30.0]_fractions.tiff").touch()
    context.post_transfer(
        tmp_path / "Position_1_[-30.0]_fractions.tiff",
        role="detector",
        required_position_files=[],
        required_strings=["fractions"],
        environment=env,
    )
    (tmp_path / "Position_2_[30.0]_fractions.tiff").touch()
    assert not context._completed_tilt_series
    context.post_transfer(
        tmp_path / "Position_2_[30.0]_fractions.tiff",
        role="detector",
        required_position_files=[],
        required_strings=["fractions"],
        environment=env,
    )
    assert len(context._tilt_series.values()) == 2
    assert context._completed_tilt_series == ["Position_1"]


@patch("requests.get")
@patch("requests.post")
def test_tomography_context_add_tomo_tilt_out_of_order(mock_post, mock_get, tmp_path):
    env = MurfeyInstanceEnvironment(
        url=urlparse("http://localhost:8000"),
        client_id=0,
        sources=[tmp_path],
        default_destinations={tmp_path: str(tmp_path)},
    )
    context = TomographyContext("tomo", tmp_path)
    (tmp_path / "Position_1_[30.0]_fractions.tiff").touch()
    context.post_transfer(
        tmp_path / "Position_1_[30.0]_fractions.tiff",
        role="detector",
        required_position_files=[],
        required_strings=["fractions"],
        environment=env,
    )
    assert context._tilt_series == {
        "Position_1": [tmp_path / "Position_1_[30.0]_fractions.tiff"]
    }
    assert (
        context._last_transferred_file == tmp_path / "Position_1_[30.0]_fractions.tiff"
    )
    (tmp_path / "Position_1_[-30.0]_fractions.tiff").touch()
    context.post_transfer(
        tmp_path / "Position_1_[-30.0]_fractions.tiff",
        role="detector",
        required_position_files=[],
        required_strings=["fractions"],
        environment=env,
    )
    assert not context._completed_tilt_series
    (tmp_path / "Position_2_[-30.0]_fractions.tiff").touch()
    context.post_transfer(
        tmp_path / "Position_2_[-30.0]_fractions.tiff",
        role="detector",
        required_position_files=[],
        required_strings=["fractions"],
        environment=env,
    )
    assert len(context._tilt_series.values()) == 2
    assert not context._completed_tilt_series
    (tmp_path / "Position_2_[30.0]_fractions.tiff").touch()
    context.post_transfer(
        tmp_path / "Position_2_[30.0]_fractions.tiff",
        role="detector",
        required_position_files=[],
        required_strings=["fractions"],
        environment=env,
    )
    assert len(context._tilt_series.values()) == 2
    assert not context._completed_tilt_series
    (tmp_path / "Position_3_[30.0]_fractions.tiff").touch()
    (tmp_path / "Position_3_[-30.0]_fractions.tiff").touch()
    context.post_transfer(
        tmp_path / "Position_3_[-30.0]_fractions.tiff",
        role="detector",
        required_position_files=[],
        required_strings=["fractions"],
        environment=env,
    )
    assert len(context._tilt_series.values()) == 3
    assert context._completed_tilt_series == ["Position_1", "Position_2"]
    context.post_transfer(
        tmp_path / "Position_3_[30.0]_fractions.tiff",
        role="detector",
        required_position_files=[],
        required_strings=["fractions"],
        environment=env,
    )
    print(context._completed_tilt_series)
    assert context._completed_tilt_series == ["Position_1", "Position_2", "Position_3"]


@patch("requests.get")
@patch("requests.post")
def test_tomography_context_add_tomo_tilt_delayed_tilt(mock_post, mock_get, tmp_path):
    env = MurfeyInstanceEnvironment(
        url=urlparse("http://localhost:8000"),
        client_id=0,
        sources=[tmp_path],
        default_destinations={tmp_path: str(tmp_path)},
    )
    context = TomographyContext("tomo", tmp_path)
    (tmp_path / "Position_1_[30.0]_fractions.tiff").touch()
    context.post_transfer(
        tmp_path / "Position_1_[30.0]_fractions.tiff",
        role="detector",
        required_position_files=[],
        required_strings=["fractions"],
        environment=env,
    )
    assert context._tilt_series == {
        "Position_1": [tmp_path / "Position_1_[30.0]_fractions.tiff"]
    }
    assert (
        context._last_transferred_file == tmp_path / "Position_1_[30.0]_fractions.tiff"
    )
    (tmp_path / "Position_1_[-30.0]_fractions.tiff").touch()
    context.post_transfer(
        tmp_path / "Position_1_[-30.0]_fractions.tiff",
        role="detector",
        required_position_files=[],
        required_strings=["fractions"],
        environment=env,
    )
    assert not context._completed_tilt_series
    (tmp_path / "Position_2_[30.0]_fractions.tiff").touch()
    context.post_transfer(
        tmp_path / "Position_2_[30.0]_fractions.tiff",
        role="detector",
        required_position_files=[],
        required_strings=["fractions"],
        environment=env,
    )
    assert len(context._tilt_series.values()) == 2
    assert context._completed_tilt_series == ["Position_1"]
    (tmp_path / "Position_2_[-30.0]_fractions.tiff").touch()
    context.post_transfer(
        tmp_path / "Position_2_[-30.0]_fractions.tiff",
        role="detector",
        required_position_files=[],
        required_strings=["fractions"],
        environment=env,
    )
    (tmp_path / "Position_1_[60.0]_fractions.tiff").touch()
    new_series = context.post_transfer(
        tmp_path / "Position_1_[60.0]_fractions.tiff",
        role="detector",
        required_position_files=[],
        required_strings=["fractions"],
        environment=env,
    )
    assert context._completed_tilt_series == ["Position_2", "Position_1"]
    assert new_series == ["Position_1"]


def test_tomography_context_initialisation_for_serialem(tmp_path):
    context = TomographyContext("serialem", tmp_path)
    assert not context._last_transferred_file
    assert context._acquisition_software == "serialem"


@patch("requests.get")
@patch("requests.post")
def test_tomography_context_add_serialem_tilt(mock_post, mock_get, tmp_path):
    env = MurfeyInstanceEnvironment(
        url=urlparse("http://localhost:8000"),
        client_id=0,
        sources=[tmp_path],
        default_destinations={tmp_path: str(tmp_path)},
    )
    context = TomographyContext("serialem", tmp_path)
    (tmp_path / "tomography_1_2_30.tiff").touch()
    context.post_transfer(
        tmp_path / "tomography_1_2_30.tiff", role="detector", environment=env
    )
    assert context._tilt_series == {"1": [tmp_path / "tomography_1_2_30.tiff"]}
    assert context._last_transferred_file == tmp_path / "tomography_1_2_30.tiff"
    (tmp_path / "tomography_1_2_-30.tiff").touch()
    context.post_transfer(
        tmp_path / "tomography_1_2_-30.tiff", role="detector", environment=env
    )
    assert context._tilt_series == {
        "1": [
            tmp_path / "tomography_1_2_30.tiff",
            tmp_path / "tomography_1_2_-30.tiff",
        ]
    }
    assert not context._completed_tilt_series
    (tmp_path / "tomography_2_2_30.tiff").touch()
    context.post_transfer(
        tmp_path / "tomography_2_2_30.tiff", role="detector", environment=env
    )
    assert len(context._tilt_series.values()) == 2
    assert context._completed_tilt_series == ["1"]


@patch("requests.get")
@patch("requests.post")
def test_tomography_context_add_serialem_decimal_tilt(mock_post, mock_get, tmp_path):
    env = MurfeyInstanceEnvironment(
        url=urlparse("http://localhost:8000"),
        client_id=0,
        sources=[tmp_path],
        default_destinations={tmp_path: str(tmp_path)},
    )
    context = TomographyContext("serialem", tmp_path)
    (tmp_path / "tomography_1_2_30.0.tiff").touch()
    context.post_transfer(
        tmp_path / "tomography_1_2_30.0.tiff", role="detector", environment=env
    )
    assert context._tilt_series == {"1": [tmp_path / "tomography_1_2_30.0.tiff"]}
    assert context._last_transferred_file == tmp_path / "tomography_1_2_30.0.tiff"
    (tmp_path / "tomography_1_2_-30.0.tiff").touch()
    context.post_transfer(
        tmp_path / "tomography_1_2_-30.0.tiff", role="detector", environment=env
    )
    assert context._tilt_series == {
        "1": [
            tmp_path / "tomography_1_2_30.0.tiff",
            tmp_path / "tomography_1_2_-30.0.tiff",
        ]
    }
    assert not context._completed_tilt_series
    (tmp_path / "tomography_2_2_30.0.tiff").touch()
    context.post_transfer(
        tmp_path / "tomography_2_2_30.0.tiff", role="detector", environment=env
    )
    assert len(context._tilt_series.values()) == 2
    assert context._completed_tilt_series == ["1"]


@patch("requests.get")
@patch("requests.post")
def test_setting_tilt_series_size_and_completion_from_mdoc_parsing(
    mock_post, mock_get, tmp_path
):
    env = MurfeyInstanceEnvironment(
        url=urlparse("http://localhost:8000"),
        client_id=0,
        sources=[tmp_path],
        default_destinations={tmp_path: str(tmp_path)},
    )
    context = TomographyContext("tomo", tmp_path)
    assert len(context._tilt_series_sizes) == 0
    context.post_transfer(
        Path(__file__).parent.parent / "util" / "test_1.mdoc",
        role="detector",
        environment=env,
        required_strings=["fractions"],
    )
    assert len(context._tilt_series_sizes) == 1
    assert context._tilt_series_sizes == {"test_1": 11}
    (tmp_path / "test_1.mdoc").touch()
    tilt = -50
    (tmp_path / f"test_1_[{tilt:.1f}]_fractions.tiff").touch()
    context.post_transfer(
        tmp_path / f"test_1_[{tilt:.1f}]_fractions.tiff",
        role="detector",
        environment=env,
        required_strings=["fractions"],
    )
    assert context._tilt_series == {
        "test_1": [tmp_path / f"test_1_[{tilt:.1f}]_fractions.tiff"]
    }
    for t in range(-40, 60, 10):
        assert not context._completed_tilt_series
        (tmp_path / f"test_1_[{t:.1f}]_fractions.tiff").touch()
        context.post_transfer(
            tmp_path / f"test_1_[{t:.1f}]_fractions.tiff",
            role="detector",
            environment=env,
            required_strings=["fractions"],
        )
    assert len(context._tilt_series["test_1"]) == 11
    assert context._completed_tilt_series == ["test_1"]



from __future__ import annotations

from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional

from pydantic import BaseModel


class Visit(BaseModel):
    start: datetime
    end: datetime
    session_id: int
    name: str
    beamline: str
    proposal_title: str

    def __repr__(self) -> str:
        return (
            "Visit("
            f"start='{self.start:%Y-%m-%d %H:%M}', "
            f"end='{self.end:%Y-%m-%d %H:%M}', "
            f"session_id='{self.session_id!r}',"
            f"name={self.name!r}, "
            f"beamline={self.beamline!r}, "
            f"proposal_title={self.proposal_title!r}"
            ")"
        )


class Sample(BaseModel):
    sample_group_id: int
    sample_id: int
    subsample_id: int
    image_path: Optional[Path]


class BLSampleImageParameters(BaseModel):
    sample_id: int
    sample_path: Path


class BLSampleParameters(BaseModel):
    sample_group_id: int


class BLSubSampleParameters(BaseModel):
    sample_id: int
    image_path: Optional[Path] = None


class ContextInfo(BaseModel):
    experiment_type: str
    acquisition_software: str


class ClientInfo(BaseModel):
    id: int


class RsyncerInfo(BaseModel):
    source: str
    destination: str
    client_id: int
    transferring: bool = True
    increment_count: int = 1
    bytes: int = 0
    increment_data_count: int = 0
    data_bytes: int = 0


class ClearanceKeys(BaseModel):
    data_collection_group: List[str]
    data_collection: List[str]
    processing_job: List[str]
    autoproc_program: List[str]


class File(BaseModel):
    name: str
    description: str
    size: int
    timestamp: float


class LifFileInfo(BaseModel):
    name: Path
    size: int
    timestamp: float
    description: str = ""


class SPAProcessingParameters(BaseModel):
    job_id: int


class ProcessFile(BaseModel):
    path: str
    description: str
    tag: str
    data_collection_id: Optional[int]
    image_number: int
    pixel_size: float
    dose_per_frame: float
    processing_job: Optional[int] = None
    autoproc_program_id: Optional[int] = None
    mc_uuid: Optional[int] = None
    voltage: float = 300
    mc_binning: int = 1
    gain_ref: Optional[str] = None
    extract_downscale: int = 1
    eer_fractionation_file: Optional[str] = None
    group_tag: Optional[str] = None


class SPAProcessFile(BaseModel):
    tag: str
    path: str
    description: str
    processing_job: Optional[int]
    data_collection_id: Optional[int]
    image_number: int
    autoproc_program_id: Optional[int]
    foil_hole_id: Optional[int]
    pixel_size: Optional[float]
    dose_per_frame: Optional[float]
    mc_binning: Optional[int] = 1
    gain_ref: Optional[str] = None
    extract_downscale: bool = True
    eer_fractionation_file: Optional[str] = None
    source: str = ""


class TiltInfo(BaseModel):
    tilt_series_tag: str
    movie_path: str
    source: str


class TiltSeriesInfo(BaseModel):
    client_id: int
    tag: str
    source: str


class TiltSeriesGroupInfo(BaseModel):
    tags: List[str]
    source: str
    tilt_series_lengths: List[int]


class CompletedTiltSeries(BaseModel):
    tilt_series: List[str]
    rsync_source: str


class SuggestedPathParameters(BaseModel):
    base_path: Path
    touch: bool = False
    extra_directory: str = ""


class DCGroupParameters(BaseModel):
    experiment_type: str
    experiment_type_id: int
    tag: str
    atlas: str = ""
    sample: Optional[int] = None


class DCParameters(BaseModel):
    voltage: float
    pixel_size_on_image: str
    experiment_type: str
    image_size_x: int
    image_size_y: int
    file_extension: str
    acquisition_software: str
    image_directory: str
    tag: str
    source: str
    magnification: float
    total_exposed_dose: Optional[float] = None
    c2aperture: Optional[float] = None
    exposure_time: Optional[float] = None
    slit_width: Optional[float] = None
    phase_plate: bool = False
    data_collection_tag: str = ""


class PreprocessingParametersTomo(BaseModel):
    dose_per_frame: float
    gain_ref: Optional[str]
    experiment_type: str
    voltage: float
    image_size_x: int
    image_size_y: int
    pixel_size_on_image: str
    motion_corr_binning: int
    manual_tilt_offset: float
    file_extension: str
    tag: str
    tilt_series_tag: str
    eer_fractionation_file: Optional[str]
    eer_fractionation: int

    class Base(BaseModel):
        dose_per_frame: float
        gain_ref: Optional[str]
        manual_tilt_offset: float
        eer_fractionation: int


class ProcessingParametersTomo(BaseModel):
    manual_tilt_offset: int
    tag: str
    tilt_series_tag: str


class ProcessingParametersSPA(BaseModel):
    tag: str
    dose_per_frame: float
    gain_ref: Optional[str]
    experiment_type: str
    voltage: float
    image_size_x: int
    image_size_y: int
    pixel_size_on_image: str
    motion_corr_binning: int
    file_extension: str
    acquisition_software: str
    use_cryolo: bool
    symmetry: str
    mask_diameter: Optional[int]
    boxsize: Optional[int]
    downscale: bool
    small_boxsize: Optional[int]
    eer_fractionation: int
    particle_diameter: Optional[float]
    magnification: Optional[int] = None
    total_exposed_dose: Optional[float] = None
    c2aperture: Optional[float] = None
    exposure_time: Optional[float] = None
    slit_width: Optional[float] = None
    phase_plate: bool = False

    class Base(BaseModel):
        dose_per_frame: Optional[float]
        gain_ref: Optional[str]
        use_cryolo: bool
        symmetry: str
        mask_diameter: Optional[int]
        boxsize: Optional[int]
        downscale: bool
        small_boxsize: Optional[int]
        eer_fractionation: int


class ProcessingJobParameters(BaseModel):
    tag: str
    recipe: str
    parameters: Dict[str, Any] = {}
    experiment_type: str = "spa"


class RegistrationMessage(BaseModel):
    registration: str
    params: Optional[Dict[str, Any]] = None


class ConnectionFileParameters(BaseModel):
    filename: str
    destinations: List[str]


class GainReference(BaseModel):
    gain_ref: Path
    rescale: bool = True
    eer: bool = False


class SessionInfo(BaseModel):
    session_id: Optional[int]
    session_name: str = ""
    rescale: bool = True


class MillingParameters(BaseModel):
    lamella_number: int
    images: List[str]
    raw_directory: str


class FractionationParameters(BaseModel):
    fractionation: int
    dose_per_frame: float
    num_frames: int = 0
    eer_path: Optional[str] = None
    fractionation_file_name: str = "eer_fractionation.txt"


class GridSquareParameters(BaseModel):
    tag: str
    x_location: Optional[float] = None
    y_location: Optional[float] = None
    x_stage_position: Optional[float] = None
    y_stage_position: Optional[float] = None
    readout_area_x: Optional[int] = None
    readout_area_y: Optional[int] = None
    thumbnail_size_x: Optional[int] = None
    thumbnail_size_y: Optional[int] = None
    pixel_size: Optional[float] = None
    image: str = ""


class FoilHoleParameters(BaseModel):
    tag: str
    name: int
    x_location: Optional[float] = None
    y_location: Optional[float] = None
    x_stage_position: Optional[float] = None
    y_stage_position: Optional[float] = None
    readout_area_x: Optional[int] = None
    readout_area_y: Optional[int] = None
    thumbnail_size_x: Optional[int] = None
    thumbnail_size_y: Optional[int] = None
    pixel_size: Optional[float] = None
    image: str = ""


class PostInfo(BaseModel):
    url: str
    data: dict



from __future__ import annotations

import logging
import os

import importlib_metadata
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles
from prometheus_client import make_asgi_app
from pydantic import BaseSettings

import murfey.server
import murfey.server.bootstrap
import murfey.server.clem.api
import murfey.server.websocket
import murfey.util.models
from murfey.server import template_files

if os.getenv("MURFEY_DEMO"):
    from murfey.server.demo_api import router
else:
    from murfey.server.api import router


log = logging.getLogger("murfey.server.main")

tags_metadata = [murfey.server.bootstrap.tag]


class Settings(BaseSettings):
    murfey_machine_configuration: str = ""


settings = Settings()

app = FastAPI(title="Murfey server", debug=True, openapi_tags=tags_metadata)

metrics_app = make_asgi_app()
app.mount("/metrics", metrics_app)

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.mount("/static", StaticFiles(directory=template_files / "static"), name="static")
app.mount("/images", StaticFiles(directory=template_files / "images"), name="images")

# Add router endpoints to the API
app.include_router(router)
app.include_router(murfey.server.bootstrap.bootstrap)
app.include_router(murfey.server.bootstrap.cygwin)
app.include_router(murfey.server.bootstrap.pypi)
app.include_router(murfey.server.bootstrap.plugins)
app.include_router(murfey.server.clem.api.router)
app.include_router(murfey.server.websocket.ws)

for r in importlib_metadata.entry_points(group="murfey.routers"):
    app.include_router(r.load())



from __future__ import annotations

import logging
import os
import threading
import time
from pathlib import Path
from typing import List

import murfey.util

log = logging.getLogger("murfey.client.wathdir_multigrid")


class MultigridDirWatcher(murfey.util.Observer):
    def __init__(
        self,
        path: str | os.PathLike,
        machine_config: dict,
        skip_existing_processing: bool = False,
    ):
        super().__init__()
        self._basepath = Path(path)
        self._skip_existing_processing = skip_existing_processing
        self._seen_dirs: List[Path] = []
        self._stopping = False
        self._machine_config = machine_config
        self.thread = threading.Thread(
            name=f"MultigridDirWatcher {self._basepath}",
            target=self._process,
            daemon=True,
        )

    def start(self):
        if self.thread.is_alive():
            raise RuntimeError("DirWatcher already running")
        log.info(f"MultigridDirWatcher thread starting for {self}")
        self.thread.start()

    def stop(self):
        log.debug("MultigridDirWatcher thread stop requested")
        self._stopping = True
        self._halt_thread = True
        if self.thread.is_alive():
            self.thread.join()
        log.debug("MultigridDirWatcher thread stop completed")

    def _process(self):
        first_loop = True
        while not self._stopping:
            for d in self._basepath.glob("*"):
                if d.name in self._machine_config["create_directories"].values():
                    if d.is_dir() and d not in self._seen_dirs:
                        self.notify(
                            d,
                            include_mid_path=False,
                            use_suggested_path=False,
                            analyse=(
                                d.name
                                in self._machine_config["analyse_created_directories"]
                            ),
                        )
                        self._seen_dirs.append(d)
                else:
                    if d.is_dir() and d not in self._seen_dirs:
                        self.notify(
                            d,
                            extra_directory=f"metadata_{d.name}",
                            include_mid_path=False,
                            analyse=True,  # not (first_loop and self._skip_existing_processing),
                            limited=True,
                        )
                        self._seen_dirs.append(d)
                    processing_started = False
                    for d02 in (d.parent.parent / d.name).glob("Images-Disc*"):
                        if d02 not in self._seen_dirs:
                            # if skip exisiting processing is set then do not process for any
                            # data directories found on the first loop
                            # this allows you to avoid triggering processing again if murfey is restarted
                            self.notify(
                                d02,
                                include_mid_path=False,
                                remove_files=True,
                                analyse=not (
                                    first_loop and self._skip_existing_processing
                                ),
                            )
                            self._seen_dirs.append(d02)
                        processing_started = d02 in self._seen_dirs
                    if not processing_started:
                        d02 = d.parent.parent / d.name
                        if (
                            d02.is_dir()
                            and d02 not in self._seen_dirs
                            and list((d.parent.parent / d.name).iterdir())
                        ):
                            self.notify(
                                d02,
                                include_mid_path=False,
                                analyse=not (
                                    first_loop and self._skip_existing_processing
                                ),
                            )
                            self._seen_dirs.append(d02)

            if first_loop:
                first_loop = False
            time.sleep(15)



from __future__ import annotations

from pathlib import Path
from typing import Tuple

from murfey.util.file_monitor import Monitor
from murfey.util.rsync import RsyncPipe


def test_a_simple_rsync_instance(tmp_path):
    (tmp_path / "from").mkdir()
    destination = tmp_path / "to"
    destination.mkdir()
    f01 = tmp_path / "from" / "file01.txt"
    f01.touch()
    rp = RsyncPipe(destination)
    rp._run_rsync(tmp_path / "from", [f01])
    assert rp._out.qsize() == 1
    transferred = rp._out.get()
    assert transferred == f01


def test_rsync_multiple_files(tmp_path):
    (tmp_path / "from").mkdir()
    destination = tmp_path / "to"
    destination.mkdir()
    f01 = tmp_path / "from" / "file01.txt"
    f01.touch()
    f02 = tmp_path / "from" / "file02.txt"
    f02.touch()
    rp = RsyncPipe(destination)
    rp._run_rsync(tmp_path / "from", [f01, f02])
    assert rp._out.qsize() == 2
    transferred = [rp._out.get()]
    transferred.append(rp._out.get())
    assert len(transferred) == 2
    assert set(transferred) == {f01, f02}


def test_rsync_a_nonexistant_file(tmp_path):
    (tmp_path / "from").mkdir()
    destination = tmp_path / "to"
    destination.mkdir()
    f01 = tmp_path / "from" / "file01.txt"
    f01.touch()
    f02 = tmp_path / "from" / "file02.txt"
    rp = RsyncPipe(destination)
    rp._run_rsync(tmp_path / "from", [f01, f02], retry=False)
    assert rp._out.qsize() == 1
    transferred = rp._out.get()
    assert transferred == f01
    assert len(rp.failed) == 1


def test_rsync_instance_on_nested_directory_structure(tmp_path):
    initial_dir = tmp_path / "from" / "nest"
    initial_dir.mkdir(parents=True)
    destination = tmp_path / "to"
    destination.mkdir()
    f01 = initial_dir / "file01.txt"
    f01.touch()
    rp = RsyncPipe(destination)
    rp._run_rsync(tmp_path / "from", [f01])
    assert rp._out.qsize() == 1
    transferred = rp._out.get()
    assert transferred == f01
    assert not len(rp.failed)
    assert (destination / "nest" / "file01.txt").exists()


def test_rsync_pipe_from_monitor(tmp_path):
    (tmp_path / "from").mkdir()
    destination = tmp_path / "to"
    destination.mkdir()
    f01 = tmp_path / "from" / "file01.txt"
    f01.touch()
    monitor = Monitor(tmp_path / "from")
    monitor.process(in_thread=True, sleep=0.1)
    rp = RsyncPipe(destination)
    monitor >> rp
    rp.process(in_thread=True)
    assert rp.thread
    monitor.stop()
    monitor.wait()
    rp.wait()
    assert (destination / "file01.txt").exists()


def test_rsync_with_additional_structure_without_changing_file_name(tmp_path):
    def _new_structure(s: Path, p: Path) -> Tuple[Path, str]:
        new_name = p.name
        new_dest = s / "extra"
        return new_dest, new_name

    (tmp_path / "from").mkdir()
    destination = tmp_path / "to"
    destination.mkdir()
    (destination / "extra").mkdir()
    f01 = tmp_path / "from" / "file01.txt"
    f01.touch()
    monitor = Monitor(tmp_path / "from")
    monitor.process(in_thread=True, sleep=0.1)
    rp = RsyncPipe(destination, destination_structure=_new_structure)
    monitor >> rp
    rp.process(in_thread=True)
    assert rp.thread
    monitor.stop()
    monitor.wait()
    rp.wait()
    assert (destination / "extra" / "file01.txt").exists()


def test_rsync_with_changed_file_name(tmp_path):
    def _new_structure(s: Path, p: Path) -> Tuple[Path, str]:
        new_name = p.name.replace("01", "05")
        return s, new_name

    (tmp_path / "from").mkdir()
    destination = tmp_path / "to"
    destination.mkdir()
    f01 = tmp_path / "from" / "file01.txt"
    f01.touch()
    monitor = Monitor(tmp_path / "from")
    monitor.process(in_thread=True, sleep=0.1)
    rp = RsyncPipe(destination, destination_structure=_new_structure)
    monitor >> rp
    rp.process(in_thread=True)
    assert rp.thread
    monitor.stop()
    monitor.wait()
    rp.wait()
    assert not (destination / "file01.txt").exists()
    assert (destination / "file05.txt").exists()



import argparse
import os

from cryptography.fernet import Fernet

from murfey.server.config import get_machine_config


def run():
    parser = argparse.ArgumentParser(description="Decrypt Murfey database password")

    parser.add_argument(nargs="?", dest="password", help="Password to decrypt")
    parser.add_argument(
        "-m",
        "--microscope",
        dest="microscope",
        type=str,
        default="",
        help="Microscope as specified in the Murfey machine configuration",
    )

    args = parser.parse_args()

    if args.microscope:
        os.environ["BEAMLINE"] = args.microscope

    machine_config = get_machine_config()
    f = Fernet(machine_config.crypto_key.encode("ascii"))
    print(f.decrypt(args.password.encode("ascii")).decode())



from __future__ import annotations

from fastapi import APIRouter

from murfey.util.lif import convert_lif_to_tiff
from murfey.util.models import LifFileInfo

# Create APIRouter class object
router = APIRouter()


# Allow function to be seen as an endpoint by the router
@router.post("/sessions/{session_id}/lif_to_tiff")
def lif_to_tiff(
    session_id: int,  # Used by the decorator
    lif_info: LifFileInfo,
):
    convert_lif_to_tiff(
        file=lif_info.name,
        root_folder="images",
    )



from __future__ import annotations

import logging
from datetime import datetime
from itertools import count
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, OrderedDict, Tuple

import requests
import xmltodict

from murfey.client.context import Context, ProcessingParameter
from murfey.client.instance_environment import (
    MovieTracker,
    MurfeyID,
    MurfeyInstanceEnvironment,
)
from murfey.util import capture_get, capture_post, get_machine_config

logger = logging.getLogger("murfey.client.contexts.spa")


class FoilHole(NamedTuple):
    session_id: int
    id: int
    grid_square_id: int
    x_location: Optional[float] = None
    y_location: Optional[float] = None
    x_stage_position: Optional[float] = None
    y_stage_position: Optional[float] = None
    readout_area_x: Optional[int] = None
    readout_area_y: Optional[int] = None
    thumbnail_size_x: Optional[int] = None
    thumbnail_size_y: Optional[int] = None
    pixel_size: Optional[float] = None
    image: str = ""


class GridSquare(NamedTuple):
    session_id: int
    id: int
    x_location: Optional[float] = None
    y_location: Optional[float] = None
    x_stage_position: Optional[float] = None
    y_stage_position: Optional[float] = None
    readout_area_x: Optional[int] = None
    readout_area_y: Optional[int] = None
    thumbnail_size_x: Optional[int] = None
    thumbnail_size_y: Optional[int] = None
    pixel_size: Optional[float] = None
    image: str = ""
    tag: str = ""


def _get_grid_square_atlas_positions(
    xml_path: Path, grid_square: str = ""
) -> Dict[str, Tuple[Optional[int], Optional[int], Optional[float], Optional[float]]]:
    with open(
        xml_path,
        "r",
    ) as dm:
        atlas_data = xmltodict.parse(dm.read())
    tile_info = atlas_data["AtlasSessionXml"]["Atlas"]["TilesEfficient"]["_items"][
        "TileXml"
    ]
    gs_pix_positions: Dict[
        str, Tuple[Optional[int], Optional[int], Optional[float], Optional[float]]
    ] = {}
    for ti in tile_info:
        try:
            nodes = ti["Nodes"]["KeyValuePairs"]
        except KeyError:
            continue
        required_key = ""
        for key in nodes.keys():
            if key.startswith("KeyValuePairOfintNodeXml"):
                required_key = key
                break
        if not required_key:
            continue
        for gs in nodes[required_key]:
            if not isinstance(gs, dict):
                continue
            if not grid_square or gs["key"] == grid_square:
                gs_pix_positions[gs["key"]] = (
                    int(float(gs["value"]["b:PositionOnTheAtlas"]["c:Center"]["d:x"])),
                    int(float(gs["value"]["b:PositionOnTheAtlas"]["c:Center"]["d:y"])),
                    float(gs["value"]["b:PositionOnTheAtlas"]["c:Physical"]["d:x"])
                    * 1e9,
                    float(gs["value"]["b:PositionOnTheAtlas"]["c:Physical"]["d:y"])
                    * 1e9,
                )
                if grid_square:
                    break
    return gs_pix_positions


def _file_transferred_to(
    environment: MurfeyInstanceEnvironment, source: Path, file_path: Path
):
    machine_config = get_machine_config(
        str(environment.url.geturl()), demo=environment.demo
    )
    if environment.visit in environment.default_destinations[source]:
        return (
            Path(machine_config.get("rsync_basepath", ""))
            / Path(environment.default_destinations[source])
            / file_path.relative_to(source)
        )
    return (
        Path(machine_config.get("rsync_basepath", ""))
        / Path(environment.default_destinations[source])
        / environment.visit
        / file_path.relative_to(source)
    )


def _grid_square_from_file(f: Path) -> int:
    for p in f.parts:
        if p.startswith("GridSquare"):
            return int(p.split("_")[1])
    raise ValueError(f"Grid square ID could not be determined from path {f}")


def _foil_hole_from_file(f: Path) -> int:
    return int(f.name.split("_")[1])


def _grid_square_metadata_file(
    f: Path, data_directories: Dict[Path, str], visit: str, grid_square: int
) -> Path:
    for dd in data_directories.keys():
        if str(f).startswith(str(dd)):
            base_dir = dd
            mid_dir = f.relative_to(dd).parent
            break
    else:
        raise ValueError(f"Could not determine grid square metadata path for {f}")
    return (
        base_dir
        / visit
        / mid_dir.parent.parent.parent
        / "Metadata"
        / f"GridSquare_{grid_square}.dm"
    )


def _grid_square_data(xml_path: Path, grid_square: int, session_id: int) -> GridSquare:
    image_paths = list(
        (xml_path.parent.parent).glob(
            f"Images-Disc*/GridSquare_{grid_square}/GridSquare_*.jpg"
        )
    )
    if image_paths:
        image_paths.sort(key=lambda x: x.stat().st_ctime)
        image_path = image_paths[0]
        with open(Path(image_path).with_suffix(".xml")) as gs_xml:
            gs_xml_data = xmltodict.parse(gs_xml.read())
        readout_area = gs_xml_data["MicroscopeImage"]["microscopeData"]["acquisition"][
            "camera"
        ]["ReadoutArea"]
        pixel_size = gs_xml_data["MicroscopeImage"]["SpatialScale"]["pixelSize"]["x"][
            "numericValue"
        ]
        full_size = (int(readout_area["a:width"]), int(readout_area["a:height"]))
        return GridSquare(
            id=grid_square,
            session_id=session_id,
            readout_area_x=full_size[0] if image_path else None,
            readout_area_y=full_size[1] if image_path else None,
            thumbnail_size_x=None,
            thumbnail_size_y=None,
            pixel_size=float(pixel_size) if image_path else None,
            image=str(image_path),
        )
    return GridSquare(id=grid_square, session_id=session_id)


def _foil_hole_data(
    xml_path: Path, foil_hole: int, grid_square: int, session_id: int
) -> FoilHole:
    with open(xml_path, "r") as xml:
        for_parsing = xml.read()
        data = xmltodict.parse(for_parsing)
    data = data["GridSquareXml"]
    serialization_array = data["TargetLocations"]["TargetLocationsEfficient"][
        "a:m_serializationArray"
    ]
    required_key = ""
    for key in serialization_array.keys():
        if key.startswith("b:KeyValuePairOfintTargetLocation"):
            required_key = key
            break
    if required_key:
        image_paths = list(
            (xml_path.parent.parent).glob(
                f"Images-Disc*/GridSquare_{grid_square}/FoilHoles/FoilHole_{foil_hole}_*.jpg"
            )
        )
        image_paths.sort(key=lambda x: x.stat().st_ctime)
        image_path: Path | str = image_paths[-1] if image_paths else ""
        if image_path:
            with open(Path(image_path).with_suffix(".xml")) as fh_xml:
                fh_xml_data = xmltodict.parse(fh_xml.read())
            readout_area = fh_xml_data["MicroscopeImage"]["microscopeData"][
                "acquisition"
            ]["camera"]["ReadoutArea"]
            pixel_size = fh_xml_data["MicroscopeImage"]["SpatialScale"]["pixelSize"][
                "x"
            ]["numericValue"]
            full_size = (int(readout_area["a:width"]), int(readout_area["a:height"]))
        for fh_block in serialization_array[required_key]:
            pix = fh_block["b:value"]["PixelCenter"]
            stage = fh_block["b:value"]["StagePosition"]
            if int(fh_block["b:key"]) == foil_hole:
                return FoilHole(
                    id=foil_hole,
                    grid_square_id=grid_square,
                    session_id=session_id,
                    x_location=float(pix["c:x"]),
                    y_location=float(pix["c:y"]),
                    x_stage_position=float(stage["c:X"]),
                    y_stage_position=float(stage["c:Y"]),
                    readout_area_x=full_size[0] if image_path else None,
                    readout_area_y=full_size[1] if image_path else None,
                    thumbnail_size_x=None,
                    thumbnail_size_y=None,
                    pixel_size=float(pixel_size) if image_path else None,
                    image=str(image_path),
                )
    logger.warning(
        f"Foil hole positions could not be determined from metadata file {xml_path} for foil hole {foil_hole}"
    )
    return FoilHole(id=foil_hole, grid_square_id=grid_square, session_id=session_id)


def _get_source(file_path: Path, environment: MurfeyInstanceEnvironment) -> Path | None:
    for s in environment.sources:
        if file_path.is_relative_to(s):
            return s
    return None


def _get_xml_list_index(key: str, xml_list: list) -> int:
    for i, elem in enumerate(xml_list):
        if elem["a:Key"] == key:
            return i
    raise ValueError(f"Key not found in XML list: {key}")


class _SPAContext(Context):
    user_params = [
        ProcessingParameter(
            "dose_per_frame",
            "Dose Per Frame [e- / Angstrom^2 / frame] (after EER grouping if relevant)",
            default=1,
        ),
        ProcessingParameter(
            "estimate_particle_diameter",
            "Use crYOLO to Estimate Particle Diameter",
            default=True,
        ),
        ProcessingParameter(
            "particle_diameter", "Particle Diameter (Angstroms)", default=None
        ),
        ProcessingParameter("use_cryolo", "Use crYOLO Autopicking", default=True),
        ProcessingParameter("symmetry", "Symmetry Group", default="C1"),
        ProcessingParameter("eer_fractionation", "EER Fractionation", default=20),
        ProcessingParameter(
            "mask_diameter", "Mask Diameter (2D classification)", default=190
        ),
        ProcessingParameter("boxsize", "Box Size", default=256),
        ProcessingParameter("downscale", "Downscale Extracted Particles", default=True),
        ProcessingParameter(
            "small_boxsize", "Downscaled Extracted Particle Size (pixels)", default=128
        ),
        ProcessingParameter("gain_ref", "Gain Reference"),
        ProcessingParameter("gain_ref_superres", "Unbinned Gain Reference"),
    ]
    metadata_params = [
        ProcessingParameter("voltage", "Voltage"),
        ProcessingParameter("image_size_x", "Image Size X"),
        ProcessingParameter("image_size_y", "Image Size Y"),
        ProcessingParameter("pixel_size_on_image", "Pixel Size"),
        ProcessingParameter("motion_corr_binning", "Motion Correction Binning"),
    ]

    def __init__(self, acquisition_software: str, basepath: Path):
        super().__init__("SPA", acquisition_software)
        self._basepath = basepath
        self._processing_job_stash: dict = {}
        self._preprocessing_triggers: dict = {}
        self._foil_holes: Dict[int, List[int]] = {}

    def gather_metadata(
        self, metadata_file: Path, environment: MurfeyInstanceEnvironment | None = None
    ):
        logger.info(f"trying to gather metadata on {metadata_file}")
        if metadata_file.suffix != ".xml":
            raise ValueError(
                f"SPA gather_metadata method expected xml file not {metadata_file.name}"
            )
        if not metadata_file.is_file():
            logger.debug(f"Metadata file {metadata_file} not found")
            return OrderedDict({})
        with open(metadata_file, "r") as xml:
            try:
                for_parsing = xml.read()
            except Exception:
                logger.warning(f"Failed to parse file {metadata_file}")
                return OrderedDict({})
            data = xmltodict.parse(for_parsing)
        magnification = 0
        num_fractions = 1
        metadata: OrderedDict = OrderedDict({})
        metadata["experiment_type"] = "SPA"
        if data.get("Acquisition"):
            metadata["voltage"] = 300
            metadata["image_size_x"] = data["Acquisition"]["Info"]["ImageSize"]["Width"]
            metadata["image_size_y"] = data["Acquisition"]["Info"]["ImageSize"][
                "Height"
            ]
            metadata["pixel_size_on_image"] = float(
                data["Acquisition"]["Info"]["SensorPixelSize"]["Height"]
            )
            metadata["magnification"] = magnification
        elif data.get("MicroscopeImage"):
            metadata["voltage"] = (
                float(
                    data["MicroscopeImage"]["microscopeData"]["gun"][
                        "AccelerationVoltage"
                    ]
                )
                / 1000
            )
            metadata["image_size_x"] = data["MicroscopeImage"]["microscopeData"][
                "acquisition"
            ]["camera"]["ReadoutArea"]["a:width"]
            metadata["image_size_y"] = data["MicroscopeImage"]["microscopeData"][
                "acquisition"
            ]["camera"]["ReadoutArea"]["a:height"]
            metadata["pixel_size_on_image"] = float(
                data["MicroscopeImage"]["SpatialScale"]["pixelSize"]["x"][
                    "numericValue"
                ]
            )
            magnification = data["MicroscopeImage"]["microscopeData"]["optics"][
                "TemMagnification"
            ]["NominalMagnification"]
            metadata["magnification"] = magnification
            try:
                dose_index = _get_xml_list_index(
                    "Dose",
                    data["MicroscopeImage"]["CustomData"]["a:KeyValueOfstringanyType"],
                )
                metadata["total_exposed_dose"] = round(
                    float(
                        data["MicroscopeImage"]["CustomData"][
                            "a:KeyValueOfstringanyType"
                        ][dose_index]["a:Value"]["#text"]
                    )
                    * (1e-20),
                    2,
                )  # convert e / m^2 to e / A^2
            except ValueError:
                metadata["total_exposed_dose"] = 1
            try:
                num_fractions = int(
                    data["MicroscopeImage"]["microscopeData"]["acquisition"]["camera"][
                        "CameraSpecificInput"
                    ]["a:KeyValueOfstringanyType"][2]["a:Value"]["b:NumberOffractions"]
                )
            except (KeyError, IndexError):
                pass
            c2_index = 3
            for i, el in enumerate(
                data["MicroscopeImage"]["CustomData"]["a:KeyValueOfstringanyType"]
            ):
                if el["a:Key"] == "Aperture[C2].Name":
                    c2_index = i
                    break
            metadata["c2aperture"] = data["MicroscopeImage"]["CustomData"][
                "a:KeyValueOfstringanyType"
            ][c2_index]["a:Value"]["#text"]
            metadata["exposure_time"] = data["MicroscopeImage"]["microscopeData"][
                "acquisition"
            ]["camera"]["ExposureTime"]
            try:
                metadata["slit_width"] = data["MicroscopeImage"]["microscopeData"][
                    "optics"
                ]["EnergyFilter"]["EnergySelectionSlitWidth"]
            except KeyError:
                metadata["slit_width"] = None
            metadata["phase_plate"] = (
                1
                if data["MicroscopeImage"]["CustomData"]["a:KeyValueOfstringanyType"][
                    11
                ]["a:Value"]["#text"]
                == "true"
                else 0
            )
        else:
            logger.warning("Metadata file format is not recognised")
            return OrderedDict({})
        binning_factor_xml = int(
            data["MicroscopeImage"]["microscopeData"]["acquisition"]["camera"][
                "Binning"
            ]["a:x"]
        )
        binning_factor = 1
        if environment:
            server_config_response = capture_get(
                f"{str(environment.url.geturl())}/machine/"
            )
            if server_config_response is None:
                return None
            server_config = server_config_response.json()
            if server_config.get("superres") and not environment.superres:
                # If camera is capable of superres and collection is in superres
                binning_factor = 2
            elif not server_config.get("superres"):
                binning_factor_xml = 2
            if magnification:
                ps_from_mag = (
                    server_config.get("calibrations", {})
                    .get("magnification", {})
                    .get(magnification)
                )
                if ps_from_mag:
                    metadata["pixel_size_on_image"] = float(ps_from_mag) * 1e-10
                    # this is a bit of a hack to cover the case when the data is binned K3
                    # then the pixel size from the magnification table will be correct but the binning_factor will be 2
                    # this is divided out later so multiply it in here to cancel
                    if server_config.get("superres") and not environment.superres:
                        metadata["pixel_size_on_image"] /= (
                            1 if binning_factor_xml == 2 else 2
                        )
                else:
                    metadata["pixel_size_on_image"] /= binning_factor
        metadata["image_size_x"] = str(int(metadata["image_size_x"]) * binning_factor)
        metadata["image_size_y"] = str(int(metadata["image_size_y"]) * binning_factor)
        metadata["motion_corr_binning"] = 1 if binning_factor_xml == 2 else 2
        if environment:
            metadata["gain_ref"] = (
                environment.data_collection_parameters.get("gain_ref")
                if environment
                and environment.data_collection_parameters.get("gain_ref")
                not in (None, "None")
                else f"data/{datetime.now().year}/{environment.visit}/processing/gain.mrc"
            )
            metadata["gain_ref_superres"] = (
                environment.data_collection_parameters.get("gain_ref_superres")
                if environment
                and environment.data_collection_parameters.get("gain_ref_superres")
                not in (None, "None")
                else f"data/{datetime.now().year}/{environment.visit}/processing/gain_superres.mrc"
            )
        else:
            metadata["gain_ref"] = None
            metadata["gain_ref_superres"] = None
        if metadata.get("total_exposed_dose"):
            metadata["dose_per_frame"] = (
                environment.data_collection_parameters.get("dose_per_frame")
                if environment
                and environment.data_collection_parameters.get("dose_per_frame")
                not in (None, "None")
                else round(metadata["total_exposed_dose"] / num_fractions, 3)
            )
        else:
            metadata["dose_per_frame"] = (
                environment.data_collection_parameters.get("dose_per_frame")
                if environment
                else None
            )

        metadata["use_cryolo"] = (
            environment.data_collection_parameters.get("use_cryolo")
            if environment
            else None
        ) or True
        metadata["symmetry"] = (
            environment.data_collection_parameters.get("symmetry")
            if environment
            else None
        ) or "C1"
        metadata["mask_diameter"] = (
            environment.data_collection_parameters.get("mask_diameter")
            if environment
            else None
        ) or 190
        metadata["boxsize"] = (
            environment.data_collection_parameters.get("boxsize")
            if environment
            else None
        ) or 256
        metadata["downscale"] = (
            environment.data_collection_parameters.get("downscale")
            if environment
            else None
        ) or True
        metadata["small_boxsize"] = (
            environment.data_collection_parameters.get("small_boxsize")
            if environment
            else None
        ) or 128
        metadata["eer_fractionation"] = (
            environment.data_collection_parameters.get("eer_fractionation")
            if environment
            else None
        ) or 20
        metadata["source"] = str(self._basepath)
        metadata["particle_diameter"] = (
            environment.data_collection_parameters.get("particle_diameter")
            if environment
            else None
        ) or 0
        metadata["estimate_particle_diameter"] = (
            environment.data_collection_parameters.get("estimate_particle_diameter")
            if environment
            else None
        ) or True
        return metadata


class SPAModularContext(_SPAContext):
    def _position_analysis(
        self,
        transferred_file: Path,
        environment: MurfeyInstanceEnvironment,
        source: Path,
        machine_config: dict,
    ) -> int:
        grid_square = _grid_square_from_file(transferred_file)
        grid_square_metadata_file = _grid_square_metadata_file(
            transferred_file,
            {Path(d): l for d, l in machine_config["data_directories"].items()},
            environment.visit,
            grid_square,
        )
        if (
            grid_square is not None
            and environment.murfey_session is not None
            and self._foil_holes.get(grid_square) is None
        ):
            self._foil_holes[grid_square] = []
            gs_pix_position: Tuple[
                Optional[int],
                Optional[int],
                Optional[float],
                Optional[float],
            ] = (None, None, None, None)
            data_collection_group = (
                requests.get(
                    f"{str(environment.url.geturl())}/sessions/{environment.murfey_session}/data_collection_groups"
                )
                .json()
                .get(str(source), {})
            )
            if data_collection_group.get("atlas"):
                visit_path = ""
                for p in transferred_file.parts:
                    if p == environment.visit:
                        break
                    visit_path += f"/{p}"
                if source in list(environment.samples.keys()):
                    local_atlas_path = (
                        Path(visit_path) / environment.samples[source].atlas
                    )
                    gs_pix_position = _get_grid_square_atlas_positions(
                        local_atlas_path,
                        grid_square=str(grid_square),
                    )[str(grid_square)]
            gs_url = f"{str(environment.url.geturl())}/sessions/{environment.murfey_session}/grid_square/{grid_square}"
            gs = _grid_square_data(
                grid_square_metadata_file,
                grid_square,
                environment.murfey_session,
            )
            metadata_source = Path(
                (
                    "/".join(source.parts[:-2])
                    + f"/{environment.visit}/"
                    + source.parts[-2]
                )[1:]
            )
            image_path = (
                _file_transferred_to(environment, metadata_source, Path(gs.image))
                if gs.image
                else ""
            )
            capture_post(
                gs_url,
                json={
                    "tag": str(source),
                    "readout_area_x": gs.readout_area_x,
                    "readout_area_y": gs.readout_area_y,
                    "thumbnail_size_x": gs.thumbnail_size_x,
                    "thumbnail_size_y": gs.thumbnail_size_y,
                    "pixel_size": gs.pixel_size,
                    "image": str(image_path),
                    "x_location": gs_pix_position[0],
                    "y_location": gs_pix_position[1],
                    "x_stage_position": gs_pix_position[2],
                    "y_stage_position": gs_pix_position[3],
                },
            )
        foil_hole = _foil_hole_from_file(transferred_file)
        if foil_hole not in self._foil_holes[grid_square]:
            fh_url = f"{str(environment.url.geturl())}/sessions/{environment.murfey_session}/grid_square/{grid_square}/foil_hole"
            if (
                grid_square_metadata_file.is_file()
                and environment.murfey_session is not None
            ):
                fh = _foil_hole_data(
                    grid_square_metadata_file,
                    foil_hole,
                    grid_square,
                    environment.murfey_session,
                )
                metadata_source = Path(
                    (
                        "/".join(source.parts[:-2])
                        + f"/{environment.visit}/"
                        + source.parts[-2]
                    )[1:]
                )
                image_path = (
                    _file_transferred_to(environment, metadata_source, Path(fh.image))
                    if fh.image
                    else ""
                )
                capture_post(
                    fh_url,
                    json={
                        "name": foil_hole,
                        "x_location": fh.x_location,
                        "y_location": fh.y_location,
                        "x_stage_position": fh.x_stage_position,
                        "y_stage_position": fh.y_stage_position,
                        "readout_area_x": fh.readout_area_x,
                        "readout_area_y": fh.readout_area_y,
                        "thumbnail_size_x": fh.thumbnail_size_x,
                        "thumbnail_size_y": fh.thumbnail_size_y,
                        "pixel_size": fh.pixel_size,
                        "tag": str(source),
                        "image": str(image_path),
                    },
                )
            else:
                capture_post(
                    fh_url,
                    json={
                        "name": foil_hole,
                        "tag": str(source),
                    },
                )
            self._foil_holes[grid_square].append(foil_hole)
        return foil_hole

    def post_transfer(
        self,
        transferred_file: Path,
        role: str = "",
        environment: MurfeyInstanceEnvironment | None = None,
        **kwargs,
    ) -> bool:
        data_suffixes = (".mrc", ".tiff", ".tif", ".eer")
        if role == "detector" and "gain" not in transferred_file.name:
            if transferred_file.suffix in data_suffixes:
                if self._acquisition_software == "epu":
                    if environment:
                        machine_config = get_machine_config(
                            str(environment.url.geturl()), demo=environment.demo
                        )
                    else:
                        machine_config = {}
                    required_strings = (
                        machine_config.get("data_required_substrings", {})
                        .get("epu", {})
                        .get(transferred_file.suffix, ["fractions"])
                    )

                    if not environment:
                        logger.warning("No environment passed in")
                        return True
                    source = _get_source(transferred_file, environment)
                    if not source:
                        logger.warning(f"No source found for file {transferred_file}")
                        return True

                    if required_strings and not any(
                        r in transferred_file.name for r in required_strings
                    ):
                        return True

                    if environment:
                        file_transferred_to = _file_transferred_to(
                            environment, source, transferred_file
                        )
                        if not environment.movie_counters.get(str(source)):
                            movie_counts_get = capture_get(
                                f"{str(environment.url.geturl())}/num_movies",
                            )
                            if movie_counts_get is not None:
                                environment.movie_counters[str(source)] = count(
                                    movie_counts_get.json().get(str(source), 0) + 1
                                )
                        environment.movies[file_transferred_to] = MovieTracker(
                            movie_number=next(environment.movie_counters[str(source)]),
                            motion_correction_uuid=next(MurfeyID),
                        )

                        eer_fractionation_file = None
                        if file_transferred_to.suffix == ".eer":
                            response = capture_post(
                                f"{str(environment.url.geturl())}/visits/{environment.visit}/eer_fractionation_file",
                                json={
                                    "eer_path": str(file_transferred_to),
                                    "fractionation": environment.data_collection_parameters[
                                        "eer_fractionation"
                                    ],
                                    "dose_per_frame": environment.data_collection_parameters[
                                        "dose_per_frame"
                                    ],
                                    "fractionation_file_name": "eer_fractionation_spa.txt",
                                },
                            )
                            if response is None:
                                return False
                            eer_fractionation_file = response.json()[
                                "eer_fractionation_file"
                            ]

                        try:
                            foil_hole: Optional[int] = self._position_analysis(
                                transferred_file, environment, source, machine_config
                            )
                        except Exception:
                            # try to continue if position information gathering fails so that movie is processed anyway
                            foil_hole = None

                        preproc_url = f"{str(environment.url.geturl())}/visits/{environment.visit}/{environment.client_id}/spa_preprocess"
                        preproc_data = {
                            "path": str(file_transferred_to),
                            "description": "",
                            "processing_job": None,
                            "data_collection_id": None,
                            "image_number": environment.movies[
                                file_transferred_to
                            ].movie_number,
                            "pixel_size": environment.data_collection_parameters.get(
                                "pixel_size_on_image"
                            ),
                            "autoproc_program_id": None,
                            "dose_per_frame": environment.data_collection_parameters.get(
                                "dose_per_frame"
                            ),
                            "mc_binning": environment.data_collection_parameters.get(
                                "motion_corr_binning", 1
                            ),
                            "gain_ref": environment.data_collection_parameters.get(
                                "gain_ref"
                            ),
                            "extract_downscale": environment.data_collection_parameters.get(
                                "downscale"
                            ),
                            "eer_fractionation_file": eer_fractionation_file,
                            "tag": str(source),
                            "foil_hole_id": foil_hole,
                        }
                        capture_post(
                            preproc_url,
                            json={
                                k: None if v == "None" else v
                                for k, v in preproc_data.items()
                            },
                        )

        return True

    def _register_data_collection(
        self,
        tag: str,
        url: str,
        data: dict,
        environment: MurfeyInstanceEnvironment,
    ):
        return

    def _register_processing_job(
        self,
        tag: str,
        environment: MurfeyInstanceEnvironment,
        parameters: Dict[str, Any] | None = None,
    ):
        return

    def _launch_spa_pipeline(
        self,
        tag: str,
        jobid: int,
        environment: MurfeyInstanceEnvironment,
        url: str = "",
    ):
        return


class SPAContext(_SPAContext):
    def _register_data_collection(
        self,
        tag: str,
        url: str,
        data: dict,
        environment: MurfeyInstanceEnvironment,
    ):
        logger.info(f"registering data collection with data {data}")
        environment.id_tag_registry["data_collection"].append(tag)
        image_directory = str(environment.default_destinations[Path(tag)])
        json = {
            "voltage": data["voltage"],
            "pixel_size_on_image": data["pixel_size_on_image"],
            "experiment_type": data["experiment_type"],
            "image_size_x": data["image_size_x"],
            "image_size_y": data["image_size_y"],
            "file_extension": data["file_extension"],
            "acquisition_software": data["acquisition_software"],
            "image_directory": image_directory,
            "tag": tag,
            "source": tag,
            "magnification": data["magnification"],
            "total_exposed_dose": data.get("total_exposed_dose"),
            "c2aperture": data.get("c2aperture"),
            "exposure_time": data.get("exposure_time"),
            "slit_width": data.get("slit_width"),
            "phase_plate": data.get("phase_plate", False),
        }
        capture_post(url, json=json)

    def post_transfer(
        self,
        transferred_file: Path,
        role: str = "",
        environment: MurfeyInstanceEnvironment | None = None,
        **kwargs,
    ) -> bool:
        return True

    def _register_processing_job(
        self,
        tag: str,
        environment: MurfeyInstanceEnvironment,
        parameters: Dict[str, Any] | None = None,
    ):
        logger.info(f"registering processing job with parameters: {parameters}")
        parameters = parameters or {}
        environment.id_tag_registry["processing_job"].append(tag)
        proc_url = f"{str(environment.url.geturl())}/visits/{environment.visit}/{environment.client_id}/register_processing_job"
        machine_config = get_machine_config(
            str(environment.url.geturl()), demo=environment.demo
        )
        image_directory = str(
            Path(machine_config.get("rsync_basepath", "."))
            / environment.default_destinations[Path(tag)]
        )
        if self._acquisition_software == "epu":
            import_images = f"{Path(image_directory).resolve()}/GridSquare*/Data/*{parameters['file_extension']}"
        else:
            import_images = (
                f"{Path(image_directory).resolve()}/*{parameters['file_extension']}"
            )
        msg: Dict[str, Any] = {
            "tag": tag,
            "recipe": "ispyb-relion",
            "parameters": {
                "acquisition_software": parameters["acquisition_software"],
                "voltage": parameters["voltage"],
                "gain_ref": parameters["gain_ref"],
                "dose_per_frame": parameters["dose_per_frame"],
                "eer_grouping": parameters["eer_fractionation"],
                "import_images": import_images,
                "angpix": float(parameters["pixel_size_on_image"]) * 1e10,
                "symmetry": parameters["symmetry"],
                "boxsize": parameters["boxsize"],
                "downscale": parameters["downscale"],
                "small_boxsize": parameters["small_boxsize"],
                "mask_diameter": parameters["mask_diameter"],
                "use_cryolo": parameters["use_cryolo"],
                "estimate_particle_diameter": parameters["estimate_particle_diameter"],
            },
        }
        if parameters["particle_diameter"]:
            msg["parameters"]["particle_diameter"] = parameters["particle_diameter"]
        capture_post(proc_url, json=msg)

    def _launch_spa_pipeline(
        self,
        tag: str,
        jobid: int,
        environment: MurfeyInstanceEnvironment,
        url: str = "",
    ):
        environment.id_tag_registry["auto_proc_program"].append(tag)
        data = {"job_id": jobid}
        capture_post(url, json=data)



from unittest.mock import Mock, create_autospec

from fastapi.testclient import TestClient
from sqlmodel import Session

from murfey.server.main import app
from murfey.server.murfey_db import murfey_db_session


def movies_return():
    return [("Supervisor_1", 2)]


expression = Mock()
expression.all = movies_return

mock_session = create_autospec(Session, instance=True)
mock_session.exec.return_value = expression


def override_murfey_db():
    try:
        db = mock_session
        yield db
    finally:
        db.close()


app.dependency_overrides[murfey_db_session] = override_murfey_db

client = TestClient(app)


def test_movie_count():
    response = client.get("/num_movies")
    assert response.status_code == 200
    assert len(mock_session.method_calls) == 2
    assert response.json() == {"Supervisor_1": 2}



"""
Contains functions that help with reading LIF files and converting them into TIFF files
as part of the cryo-CLEM workflow.
"""

from __future__ import annotations

import logging
from pathlib import Path
from typing import Generator, List, Optional, Tuple
from xml.etree import ElementTree as ET

# import matplotlib.pyplot as plt
import numpy as np
from readlif.reader import LifFile
from tifffile import imwrite

# Create logger object to output messages with
logger = logging.getLogger("murfey.util.lif")


def sanitise(in_string: str) -> str:
    return in_string.replace("\r\n", "").replace("\n", "")


def get_xml_metadata(
    file: LifFile,
    save_xml: Optional[Path] = None,
) -> ET.Element:
    """
    Extracts and returns the file metadata as a formatted XML tree, and optionally
    saves it as an XML file to the specified file path.
    """

    # Use readlif function to get XML metadata
    xml_root: ET.Element = file.xml_root  # This one for navigating
    xml_tree = ET.ElementTree(xml_root)  # This one for saving

    # Skip saving the metadata if save_xml not provided
    if save_xml:
        xml_file = str(save_xml)  # Convert Path to string
        ET.indent(xml_tree, "  ")  # Format with proper indentation
        xml_tree.write(xml_file, encoding="utf-8")  # Save
        logger.info(f"File metadata saved to {xml_file}")

    return xml_root


def get_image_elements(root: ET.Element) -> List[ET.Element]:
    """
    Searches the XML metadata recursively to find the nodes tagged as "Element" that
    have image-related tags. Some LIF datasets have layers of nested elements, so a
    recursive approach is needed to avoid certain datasets breaking it.
    """

    # Nested function which generates list of elements
    def _find_elements_recursively(
        node: ET.Element,
    ) -> Generator[ET.Element, None, None]:

        # Find items labelled "Element" under current node
        elem_list = node.findall("./Children/Element")
        if len(elem_list) < 1:  # Try alternative path for top-level of XML tree
            elem_list = node.findall("./Element")

        # Recursively search for items tagged as Element under child branches
        for elem in elem_list:
            yield elem
            new_node = elem  # New starting point for the search
            new_elem_list = _find_elements_recursively(new_node)  # Call self
            for new_elem in new_elem_list:
                yield new_elem

    # Get initial list of elements
    elem_list = list(_find_elements_recursively(root))

    # Keep only the element nodes that have image-related tags
    elem_list = [elem for elem in elem_list if elem.find("./Data/Image")]

    return elem_list


def raise_bit_depth_error(bit_depth: int):
    """
    Raises an exception if the bit depth value provided is not one that NumPy can
    handle.
    """

    raise Exception(
        "The channel bit depth provided is not compatible with Numpy. "
        "Only 8, 16, 32, and 64-bit channel depths are allowed. "
        f"Current bit depth: {bit_depth}"
    )


def change_bit_depth(
    array: np.ndarray,
    bit_depth: int,
) -> np.ndarray:
    """
    Change the bit depth of the array without changing the values (barring rounding).
    """

    # Use shorter terms
    arr = array

    # NumPy defaults to float64; revert back to unsigned int
    if bit_depth == 8:
        arr = arr.astype(np.uint8)
    elif bit_depth == 16:
        arr = arr.astype(np.uint16)
    elif bit_depth == 32:
        arr = arr.astype(np.uint32)
    elif bit_depth == 64:
        arr = arr.astype(np.uint64)
    else:
        raise_bit_depth_error(bit_depth)
    return arr


def rescale_across_channel(
    array: np.ndarray,
    bit_depth: int,
    percentile_range: Optional[Tuple[float, float]],  # Lower and upper percentiles
    round_to: Optional[int] = 16,  # Round bounds to reasonably granular power of 2
) -> np.ndarray:
    """
    Checks the range of pixel values occupied by the data, then rescales it across the
    channel's bit depth.
    """

    # Check that bit depth is valid before processing even begins
    if not any(bit_depth == b for b in [8, 16, 32, 64]):
        raise_bit_depth_error(bit_depth)
    else:
        pass  # Proceed to rest of function

    # Use shorter variable names
    arr = array

    # Check if percentiles are provided
    if not percentile_range:
        logger.warning("No percentile range provided. Returning original array.")
    else:
        # Use shorter variables
        p_lo = percentile_range[0]
        p_up = percentile_range[1]

        # Calculate lower and upper bounds
        b_lo = np.floor(np.percentile(arr, p_lo) / round_to) * round_to
        b_up = (np.ceil(np.percentile(arr, p_up) / round_to) * round_to) - 1

        # Rescale across channel bit depth
        arr[arr < b_lo] = b_lo  # Overwrite lower outliers
        arr[arr > b_up] = b_up  # Overwrite upper outliers
        arr = arr - b_lo  # Shift lower bound to zero
        arr = (arr / (b_up - b_lo)) * (
            2**bit_depth - 1
        )  # Ensure data points don't exceed bit depth (max bit is 2**n - 1)

        # This step probably not needed
        # Overwrite values that exceed current channel bit depth
        # arr[arr >= (2**bit_depth - 1)] = (
        #     2**bit_depth - 1
        # )

        # Change bit depth back to initial one
        arr = change_bit_depth(arr, bit_depth)

    return arr


def rescale_to_bit_depth(
    array: np.ndarray,
    initial_bit_depth: int,
    target_bit_depth: int,
) -> Tuple[np.ndarray, int]:
    """
    Rescales the pixel values of the array to fit within the desired channel bit depth.
    Returns the array and the target bit depth as a tuple.
    """

    # Use shorter names for variables
    arr = array
    bit_init = initial_bit_depth
    bit_final = target_bit_depth

    # Check that target bit depth is allowed
    if not any(bit_final == b for b in [8, 16, 32, 64]):
        raise_bit_depth_error(bit_final)

    # Rescale (DIVIDE BEFORE MULTIPLY)
    arr = (arr / (2**bit_init - 1)) * (2**bit_final - 1)

    # This step probably not needed anymore
    # Overwrite pixels that exceed channel bit depth
    # arr[arr >= (2**bit_final - 1)] = 2**bit_final - 1

    # Change to correct unsigned integer type
    arr = change_bit_depth(arr, bit_final)

    return arr, bit_final


def convert_lif_to_tiff(
    file: Path,
    root_folder: str,  # The name of the folder to be treated as the root
):
    """
    Takes a LIF file, extracts its metadata as an XML tree, then parses through the
    sub-images stored inside it, saving each channel in the sub-image as a separate
    image stack. It uses information stored in the XML metadata to name the individual
    image stacks.

    FOLDER STRUCTURE
    ================
    Here is the folder structure of a typical DLS eBIC experiment session, with the
    folders created as part of the workflow shown as well.

    parent_folder   <- Session ID
    |_ processing   <- Created by DLS; will be archived; not used
    |_ spool        <- Created by DLS; for confidential work; not used
    |_ tmp          <- Created by DLS; for intermediate files; not used
    |_ xml          <- Created by DLS; not used
    |_ images       <- Created by us; raw data stored here
    |  |_ sample_name   <- Folders for samples
    |     |_ lif files  <- LIF files of specific sample
    |     |_ metadata   <- Created by us; Save raw XML metadata file here
    |_ processed    <- ARCHIVED BY DLS; Created by us
    |  |_ sample_name
    |     |_ lif_file_names     <- Folders for data from the same LIF file
    |        |_ sub_image       <- Folders for individual sub-images
    |           |_ tiffs        <- Save channels as individual image stacks
    |           |_ metadata     <- Individual XML files saved here (not yet implemented)
    """

    # Set up new directories
    # Identify the root directory
    root_parts = []
    for p in file.parts:  # Iterate through parts until hitting root folder
        if p.lower() == root_folder.lower():  # Eliminate case-sensitivity
            break
        root_parts.append(p)
    else:
        logger.error(
            f"Subpath {sanitise(root_folder)} was not found in image path {sanitise(str(file))}"
        )
        return None
    root_dir = Path("/".join(root_parts))  # Session ID folder

    # Get remaining path to file from root folder
    child_parts = []  # Path from the root directory to the file
    for p in reversed(file.parts):
        # Append everything up until the root directory
        if p == root_folder:
            break
        child_parts.append(p)
    else:
        logger.error(
            f"Subpath {sanitise(root_folder)} was not found in image path {sanitise(str(file))}"
        )
    child_path = Path(
        "/".join(reversed(child_parts))
    )  # Reverse it to get the right order

    # Create directory to store processed files in
    process_dir = (
        root_dir / "processed" / child_path.stem
    )  # Replace root folder with "processed"

    # Save raw XML metadata here
    raw_xml_dir = file.parent / "metadata"

    # Create new folders if not already present
    for folder in [process_dir, raw_xml_dir]:
        if not folder.exists():
            folder.mkdir(parents=True)
            logger.info(f"Created {folder}")
        else:
            logger.info(f"{folder} already exists")

    # Load LIF file as a LifFile class
    logger.info(f"Loading {sanitise(file.name)}")
    lif_file = LifFile(str(file))  # Stack of scenes
    scene_list = list(lif_file.get_iter_image())  # List of scene names

    # Save original metadata as XML tree
    logger.info("Extracting image metadata")
    xml_root = get_xml_metadata(
        file=lif_file,
        save_xml=raw_xml_dir.joinpath(file.stem + ".xml"),
    )

    # Recursively generate list of metadata-containing elements
    elem_list = get_image_elements(xml_root)

    # Check that elements match number of images
    if not len(elem_list) == len(scene_list):
        raise Exception(
            "Error matching metadata list to list of sub-images. \n"
            # Show what went wrong
            f"Metadata entries: {len(elem_list)} \n"
            f"Sub-images: {len(scene_list)}"
        )
    else:
        pass  # Carry on

    # Iterate through scenes
    logger.info("Examining sub-images")
    for i in range(len(scene_list)):

        # Load image
        img = lif_file.get_image(i)  # Set sub-image

        # Get name of sub-image
        elem = elem_list[i]  # Select corresponding element
        img_name = elem.attrib["Name"]  # Get sub-image name
        logger.info(f"Examining {img_name}")

        # Load relevant metadata (channels, dimensions, timestamps etc.)
        channels = elem.findall(
            "Data/Image/ImageDescription/Channels/ChannelDescription"
        )
        # Might be useful in the future
        # timestamps = elem.find("Data/Image/TimeStampList")
        # dimensions = elem.findall(
        #     "Data/Image/ImageDescription/Dimensions/DimensionDescription"
        # )

        # Create save dirs for TIFF files and their metadata
        img_dir = process_dir / img_name
        img_xml_dir = img_dir / "metadata"
        for folder in [img_dir, img_xml_dir]:
            if not folder.exists():
                folder.mkdir(parents=True)
                logger.info(f"Created {folder}")
            else:
                logger.info(f"{folder} already exists")

        # Parijat wants the images in 8-bit; scale down from 16-bit
        # Save channels as individual TIFFs
        for c in range(len(list(img.get_iter_c()))):
            # Get color
            color = channels[c].attrib["LUTName"]
            logger.info(f"Examining the {color.lower()} channel")

            # Extract image data to array
            logger.info("Loading image stack")
            arr: np.ndarray = []  # Array to store frames in
            # Iterate over slices
            for z in range(len(list(img.get_iter_z()))):
                frame = img.get_frame(z=z, t=0, c=c)  # PIL object; array-like
                arr.append(frame)
            arr = np.array(arr)  # Make independent copy of this array

            # Initial rescaling if bit depth not 8, 16, 32, or 64-bit
            bit_depth = img.bit_depth[c]  # Initial bit depth
            if not any(bit_depth == b for b in [8, 16, 32, 64]):
                logger.info("Bit depth non-standard, converting to 16-bit")
                arr, bit_depth = rescale_to_bit_depth(
                    array=arr, initial_bit_depth=bit_depth, target_bit_depth=16
                )
            else:
                pass

            # Rescale intensity values for fluorescent channels
            if any(
                color.lower() in key for key in ["red", "green"]
            ):  # Eliminate case-sensitivity
                logger.info(f"Rescaling {color.lower()} channel across channel depth")
                arr = rescale_across_channel(
                    array=arr,
                    bit_depth=bit_depth,
                    percentile_range=(0.5, 99.5),
                    round_to=16,
                )

            # Convert to 8-bit
            logger.info("Converting to 8-bit image")
            arr, bit_depth = rescale_to_bit_depth(
                arr, initial_bit_depth=bit_depth, target_bit_depth=8
            )

            # Get x, y, and z scales
            # Get resolution (pixels per um)
            x_res = img.scale[0]
            y_res = img.scale[1]

            # Might be used in future versions
            # Get pixel size (um per pixel)
            # x_scale = 1 / x_res
            # y_scale = 1 / y_res

            # Check that depth axis exists
            if not img.scale[2]:
                z_res: float = 0
                z_scale: float = 0  # Avoid divide by zero errors
            else:
                z_res = img.scale[2]  # Pixels per um
                z_scale = 1 / z_res  # um per pixel

            # Generate slice labels
            image_labels = [f"{f}" for f in range(len(list(img.get_iter_z())))]

            # Save as a greyscale TIFF
            save_name = img_dir.joinpath(color + ".tiff")
            logger.info(f"Saving {color.lower()} image as {save_name}")
            imwrite(
                save_name,
                arr,
                imagej=True,  # ImageJ comppatible
                photometric="minisblack",  # Grayscale image
                shape=np.shape(arr),
                dtype=arr.dtype,
                resolution=(x_res * 10**6 / 10**6, y_res * 10**6 / 10**6),
                metadata={
                    "spacing": z_scale,
                    "unit": "micron",
                    "axes": "ZYX",
                    "Labels": image_labels,
                },
            )
    return None



"""
Add a CLEM class context (refer to src/murfey/client/contexts/fib.py)

Provide a post transfer function to pass on:
- File path on the DLS file system
- Session ID (src/murfey/client/instance_environment.py)
"""

# import requests
import logging
from datetime import datetime
from pathlib import Path
from typing import Optional

from murfey.client.context import Context
from murfey.client.instance_environment import MurfeyInstanceEnvironment
from murfey.util import capture_post, get_machine_config

# Create logger object
logger = logging.getLogger("murfey.client.contexts.clem")


def _file_transferred_to(
    environment: MurfeyInstanceEnvironment, source: Path, file_path: Path
) -> Optional[Path]:
    machine_config = get_machine_config(
        str(environment.url.geturl()), demo=environment.demo
    )
    return (
        Path(machine_config.get("rsync_basepath", ""))
        / machine_config.get("rsync_module", "data")
        / str(datetime.now().year)
        / source.name
        / file_path.relative_to(source)
    )


def _get_source(
    file_path: Path, environment: MurfeyInstanceEnvironment
) -> Optional[Path]:
    for s in environment.sources:
        if file_path.is_relative_to(s):
            return s
    return None


class CLEMContext(Context):
    def __init__(self, acquisition_software: str, basepath: Path):
        super().__init__("CLEM", acquisition_software)
        self._basepath = basepath

    def post_transfer(
        self,
        transferred_file: Path,
        role: str = "",
        environment: Optional[MurfeyInstanceEnvironment] = None,
        **kwargs,
    ) -> bool:
        super().post_transfer(
            transferred_file, role=role, environment=environment, **kwargs
        )
        # Check if file is a LIF file
        if transferred_file.suffix == ".lif":
            # Type checking to satisfy MyPy
            if not environment:
                logger.warning("No environment passed in")
                return True

            # Location of the file on the client PC
            source = _get_source(transferred_file, environment)
            # Type checking to satisfy MyPy
            if not source:
                logger.warning(f"No source found for file {transferred_file}")
                return True

            # Construct the URL for the Murfey server to communicate with
            url = f"{str(environment.url.geturl())}/sessions/{environment.murfey_session}/lif_to_tiff"
            # Type checking to satisfy MyPy
            if not url:
                logger.warning("No url found for the environment")
                return True

            # Get the Path on the DLS file system
            file_path = _file_transferred_to(
                environment=environment,
                source=source,
                file_path=transferred_file,
            )

            # Get the file size and timestamp from transferred_file
            # Client PC cannot see file_path; that is for server PC

            # Post the message and logs it if there's an error
            capture_post(
                url,
                json={
                    "name": str(file_path),
                    "size": transferred_file.stat().st_size,
                    "timestamp": transferred_file.stat().st_ctime,
                    "description": "",
                },
            )
        return True



import argparse
import os

from murfey.util.db import clear, setup


def run():
    parser = argparse.ArgumentParser(
        description="Generate the necessary tables for the Murfey database"
    )

    parser.add_argument(
        "--no-clear",
        dest="clear",
        default=True,
        action="store_false",
        help="Do not clear current database tables before creating specified tables",
    )
    parser.add_argument(
        "-m",
        "--microscope",
        dest="microscope",
        type=str,
        default="",
        help="Microscope as specified in the Murfey machine configuration",
    )

    args = parser.parse_args()
    if args.microscope:
        os.environ["BEAMLINE"] = args.microscope

    from murfey.server.murfey_db import url

    if args.clear:
        clear(url())
    setup(url())
